# -*- coding: utf-8 -*-
from .asset.asset import OxaigenAsset
from .authentication.authentication import OxaigenAuthentication
from .storage.data_storage import OxaigenDataStorage


class Oxaigen:
    """
    Oxaigen SDK Class
    """

    def __init__(self):
        self.auth = OxaigenAuthentication()
        self.asset = OxaigenAsset()
        self.storage = OxaigenDataStorage()

    def login(self):
        """
        Interactive function to log in.

        This function prompts the user for a username and password
        (hidden for security reasons). It then authenticates the user
        and generates a token, saving it to a `token.json` file in the
        `.oxaigen` directory in the user's home directory.

        Returns:
            None
        """
        self.auth.login()



# -*- coding: utf-8 -*-
import os
import pickle
import shutil
import logging
import requests
from datetime import datetime
from typing import List, Optional, Dict, Any

from ..util.api import run_api_query
from ..util.exception import OxaigenSDKException, OxaigenApiException
from ..config import USER_ROOT_DIR

S3_ASSET_DOWNLOAD_LINK = "getS3AssetDownloadLink"
FILE_NAME = "fileName"
FILE_SIZE = "fileSize"
FILE_TYPE = "fileType"
DOWNLOAD_URL = "downloadUrl"
LAST_MODIFIED = "lastModified"


class OxaigenAsset:
    """
    Oxaigen Asset class
    """

    def __init__(self):
        super().__init__()

    def get_asset(
            self,
            asset_key: List[str],
            file_path: Optional[List[str]] = None,
            run_id: Optional[str] = None,
            use_cache: bool = True
    ) -> Optional[Any]:
        """
        Download an Asset from the Oxaigen Orchestration data plane

        Args:
            asset_key (List[str]): The key identifying the asset.
            file_path (Optional[List[str]], optional): The path to save the downloaded asset. Defaults to None.
            run_id (Optional[str], optional): The ID of the run. Defaults to None.
            use_cache (bool, optional): Whether to use the cache. Defaults to True.

        Returns:
            Optional[Any]: The downloaded asset.
        """
        if use_cache:
            return self._get_asset_from_cache(asset_key=asset_key, file_path=file_path, run_id=run_id)
        else:
            return self.download_asset(asset_key=asset_key, file_path=file_path, run_id=run_id)

    def _get_asset_from_cache(
            self,
            asset_key: List[str],
            file_path: Optional[List[str]],
            run_id: Optional[str] = None
    ) -> Optional[Any]:
        """
        Function to get an Asset from the DataPlane from cache (local file storage)

        Args:
            asset_key (List[str]): The key identifying the asset.
            file_path (Optional[List[str]]): The path to save the downloaded asset.
            run_id (Optional[str], optional): The ID of the run. Defaults to None.

        Returns:
            Optional[Any]: The downloaded asset.
        """
        if not file_path:
            file_path = asset_key

        data = self._get_download_asset_link(asset_key=asset_key, run_id=run_id)
        file_name = data[S3_ASSET_DOWNLOAD_LINK][FILE_NAME]
        clean_file_name = os.path.basename(file_name)
        full_path = os.path.join(USER_ROOT_DIR, *file_path)
        file = os.path.join(full_path, clean_file_name)

        asset = self.get_asset_from_file(file=file)

        # fallback to fresh download
        if asset is None:
            return self.download_asset(asset_key=asset_key, file_path=file_path, run_id=run_id)

        return asset

    def download_asset(
            self,
            asset_key: List[str],
            file_path: Optional[List[str]] = None,
            run_id: Optional[str] = None
    ) -> Optional[Any]:
        """
        Download an Asset from the Oxaigen Orchestration data plane

        Args:
            asset_key (List[str]): The key identifying the asset.
            file_path (Optional[List[str]], optional): The path to save the downloaded asset. Defaults to None.
            run_id (Optional[str], optional): The ID of the run. Defaults to None.

        Returns:
            Optional[Any]: The downloaded asset.
        """
        if not file_path:
            file_path = asset_key
        try:
            data = self._get_download_asset_link(asset_key=asset_key, run_id=run_id)

            # Check if the asset was not found
            if 'getS3AssetDownloadLink' in data and data['getS3AssetDownloadLink']['__typename'] == 'AssetNotFound':
                error_message = data['getS3AssetDownloadLink']['errorMessage']
                raise OxaigenSDKException(message=f"Asset not found: {error_message}")

            # Check if there's an error message
            if 'errorMessage' in data:
                error_message = data['errorMessage']
                raise OxaigenSDKException(message=f"Error occurred: {error_message}")

            url = data[S3_ASSET_DOWNLOAD_LINK][DOWNLOAD_URL]
            file_name = data[S3_ASSET_DOWNLOAD_LINK][FILE_NAME]
            file_size = data[S3_ASSET_DOWNLOAD_LINK][FILE_SIZE]
            clean_file_name = os.path.basename(file_name)

            logging.info(msg=f"Downloading file {clean_file_name} ({file_size}) into this directory: {file_path}")
            self._download_asset_file(url=url, file_path=file_path, file_name=clean_file_name)

            full_path = os.path.join(USER_ROOT_DIR, *file_path)
            file = os.path.join(full_path, clean_file_name)

            asset = self.get_asset_from_file(file=file)
            if asset is None:
                raise OxaigenSDKException(message=f"Could not open downloaded asset, try again")

            return asset

        except (OxaigenSDKException, OxaigenApiException):
            # Rethrow the exception
            raise

        except Exception as e:
            # Catch all other exceptions
            raise OxaigenSDKException(message=f"Could not download asset, Error: {str(e)}")

    @staticmethod
    def _get_download_asset_link(
            asset_key: List[str],
            run_id: Optional[str]
    ) -> Optional[Dict]:
        """
        Function to get a temporary download link from the Oxaigen Client API to a file in the Oxaigen data plane

        Args:
            asset_key (List[str]): The key identifying the asset.
            run_id (Optional[str]): The ID of the run.

        Returns:
            Optional[Dict]: Temporary download link data.
        """
        if not run_id:
            run_id = ""

        query = """
    query GetDownloadLink($assetKey: [String!]!, $runId: String = "") {
      getS3AssetDownloadLink(assetKey: $assetKey, runId: $runId) {
        ... on AssetDownloadLink {
          __typename
          downloadUrl
          fileSize
          fileType
          fileName
          lastModified
        }
        ... on AssetNotFound {
          __typename
          errorMessage
        }
        ... on AssetDownloadError {
          __typename
          errorMessage
        }
        ... on AuthenticationError {
          __typename
          errorMessage
        }
        ... on AuthorizationError {
          __typename
          errorMessage
        }
      }
    }
                """

        variables = {
            "assetKey": asset_key,
            "runId": run_id
        }

        return run_api_query(query=query, variables=variables)

    @staticmethod
    def _download_asset_file(url: str, file_path: List[str], file_name: str):
        """
         Function to download a file from bucket storage system using an S3 generated URL (with one-time-use credentials)

         Args:
             url (str): The URL to download the file from.
             file_path (List[str]): The path to save the downloaded asset.
             file_name (str): The name of the file to be downloaded.
         """
        # Create directory with path prefix
        full_path = os.path.join(USER_ROOT_DIR, *file_path)
        os.makedirs(full_path, exist_ok=True)
        full_file_path = os.path.join(full_path, file_name)


        # Check if the file exists before moving it
        if os.path.exists(full_file_path):
            if os.path.exists(full_path):
                file_size_mb = os.path.getsize(full_file_path) / (1024 * 1024)
                if file_size_mb < 200:
                    backup_name = f"{file_name}__backup_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
                    backup_path = os.path.join(full_path, backup_name)
                    shutil.move(full_file_path, backup_path)
                else:
                    logging.warning(
                        msg=f"Existing asset size ({str(file_size_mb)}) too large (>200mb), will not create backup")

        # Download the file
        response = requests.get(url, stream=True)
        with open(full_file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)

        logging.info(f"File {str(full_file_path)} downloaded!")

    @staticmethod
    def get_asset_from_file(file: str) -> Optional[Any]:
        """
        get an Asset from the Oxaigen Orchestration data plane from the internal User Cache

        Args:
            file (str): The path to the asset file.

        Returns:
            Optional[Any]: The asset data.
        """
        try:
            with open(file, 'rb') as f:
                data = pickle.load(f)
            return data
        except Exception as e:
            logging.warning(msg=f"Error reading asset file ({str(file)}) from disk. Error: {str(e)}")
            return None



# -*- coding: utf-8 -*-
import logging

import pandas as pd
from dagster import ConfigurableIOManager
from dagster import OpExecutionContext

class OxaigenDbIOManager(ConfigurableIOManager):
    """Oxaigen IOManager to handle loading the contents of tables as pandas DataFrames.

    Does not handle cases where data is written to different schemas for different outputs, and
    uses the name of the asset key as the table name.
    """

    con_string: str

    def handle_output(self, context: OpExecutionContext, obj):
        if isinstance(obj, pd.DataFrame):
            # write df to table
            asset_key = None
            if context is not None:
                try:
                    asset_key = context.asset_key.path[-1]
                except Exception as e:
                    print("NO ASSET KEY DEFINED, CANNOT UPLOAD TO DATABASE")
                    logging.error(e)
            if asset_key:
                print("PANDAS DATAFRAME IS BEING UPLOADED TO DATABASE!")
                obj.to_sql(name=asset_key, con=self._con, if_exists="replace")  # noqa
        elif obj is None:
            # dbt has already written the data to this table
            pass
        else:
            context.log.info(f"Unsupported object type {type(obj)} for DbIOManager, doing nothing.")
            pass

    def load_input(self, context) -> pd.DataFrame:
        """Load the contents of a table as a pandas DataFrame."""
        try:
            model_name = context.asset_key.path[-1]
            schema_name = context.asset_key.path[-2].lower()
            return pd.read_sql(f"""SELECT * FROM "{schema_name}"."{model_name}";""", con=self.con_string)
        except Exception as error:
            print(context)
            print(error)



# -*- coding: utf-8 -*-
import os
from typing import Optional

from .constant import (
    TOKEN_FILE_NAME,
    OXAIGEN_CONFIG_DIR,
)

CLIENT_SERVER_IP_ENV_VAR = "CLIENT_SERVER_IP"
CLIENT_SERVER_PORT_ENV_VAR = "CLIENT_SERVER_PORT"
ROOT_DIR_ENV_VAR = "ROOT_DIR"
JUPYTERHUB_ROOT_DIR_ENV_VAR = "JUPYTERHUB_ROOT_DIR"


def get_env_var_with_fallback(primary: str, fallback: str, default: Optional[str] = None) -> Optional[str]:
    """
    Retrieves an environment variable value with a fallback to another variable if the primary one is not set.

    Args:
        primary (str): The primary environment variable to retrieve.
        fallback (str): The fallback environment variable to use if the primary one is not set.
        default (Optional[str], optional): Default value to return if both primary and fallback variables are not set.
            Defaults to None.

    Returns:
        Optional[str]: The value of the primary environment variable if set, otherwise the value of the fallback
            environment variable, or the default value if neither are set.
    """
    return os.getenv(primary) or os.getenv(fallback, default)



# General configuration
USER_ROOT_DIR = get_env_var_with_fallback(ROOT_DIR_ENV_VAR, JUPYTERHUB_ROOT_DIR_ENV_VAR)

# Authentication configuration
TOKEN_FILE_PATH = os.path.join(f"{USER_ROOT_DIR}/{OXAIGEN_CONFIG_DIR}", TOKEN_FILE_NAME)

# API configuration
API_ENDPOINT = f"http://{os.environ.get(CLIENT_SERVER_IP_ENV_VAR)}:{os.environ.get(CLIENT_SERVER_PORT_ENV_VAR)}/v1/graphql"

# Data Storage configuration
S3_ENDPOINT = get_env_var_with_fallback('S3_ENDPOINT', 'JUPYTERLAB_S3_ENDPOINT')
S3_ACCESS_KEY_ID = get_env_var_with_fallback('S3_ACCESS_KEY_ID', 'JUPYTERLAB_S3_ACCESS_KEY_ID')
S3_SECRET_ACCESS_KEY = get_env_var_with_fallback('S3_SECRET_ACCESS_KEY', 'JUPYTERLAB_S3_SECRET_ACCESS_KEY')
S3_SSL = get_env_var_with_fallback('S3_SSL', 'JUPYTERLAB_S3_SSL', 'True').lower() == 'true'
S3_REGION = get_env_var_with_fallback('S3_REGION', 'JUPYTERLAB_S3_REGION')



# -*- coding: utf-8 -*-
from typing import Any, Mapping, Optional
from dagster_dbt import DagsterDbtTranslator
from dagster import AssetKey


class _TemplateDagsterDbtTranslator(DagsterDbtTranslator):
    """
    THIS CLASS MUST NEVER BE USED DIRECTLY!

    use create_custom_dbt_translator_class() instead as it also
    inserts the required get_asset_key() class method.
    """

    @classmethod
    def get_group_name(cls, dbt_resource_props: Mapping[str, Any]) -> Optional[str]:
        return "datawarehouse"

    @classmethod
    def get_metadata(cls, dbt_resource_props: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Custom function to set EXTRA metadata key-value pairs for a DBT asset
        """
        return {"Oxaigen": "DBT resource"}

    @classmethod
    def get_description(cls, dbt_resource_props: Mapping[str, Any]) -> str:
        """
        Custom function to set the description of the DBT asset which by default incudes the raw SQL code.
        This behaviour can be adapted below
        """
        custom_description = dbt_resource_props["name"] + "\n" + dbt_resource_props[
            "raw_code"]
        return custom_description


def create_oxaigen_dbt_translator_class(
        custom_prefix: str,
        source_tags_as_asset_prefix: bool = True,
        group_prefix: Optional[str] = None,
        source_group_prefix: Optional[str] = None
):
    class OxaigenDagsterDbtTranslator(_TemplateDagsterDbtTranslator):
        @classmethod
        def get_asset_key(cls, dbt_resource_props: Mapping[str, Any]) -> AssetKey:
            asset_key = super().get_asset_key(dbt_resource_props)

            if group_prefix:
                asset_key = asset_key.with_prefix(group_prefix)

            if dbt_resource_props["resource_type"] == "source":

                if source_tags_as_asset_prefix:
                    dbt_tags = dbt_resource_props['tags']
                    if dbt_tags:
                        for item in dbt_tags:
                            asset_key = asset_key.with_prefix(item)

                if source_group_prefix:
                    asset_key = asset_key.with_prefix(source_group_prefix)

            else:
                if dbt_resource_props.get("alias"):
                    asset_key[0][-1] = dbt_resource_props["alias"]
                else:
                    last_string = asset_key[0][-1]
                    if last_string.startswith("prod_"):
                        last_string = last_string.replace("prod_", "", 1)
                    new_string = last_string.replace("_", " ").title().replace(" ", "")
                    asset_key[0][-1] = new_string
                asset_key = asset_key.with_prefix(f"{custom_prefix}")

            return asset_key

    return OxaigenDagsterDbtTranslator()



# -*- coding: utf-8 -*-
from .src.main import Oxaigen
from .src import test

# from .src.orchestration import (
#     create_oxaigen_dbt_translator_class,
#     OxaigenDbIOManager
# )


def validate_oxaigen_sdk_import():
    """
    Validates the import of the Oxaigen SDK.

    This function checks if the Oxaigen SDK is properly imported and prints a success message if it is.

    Returns:
        None
    """
    print("Success!")

